########################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#

import logging
import os
import shutil
import stat
import urlgrabber
import urllib
import urllib2
import urlparse
grabber = urlgrabber.grabber.URLGrabber()

import Errors
import Utils.Proxy
from Utils import PathUtils
# can we get rid of this stupidity and just move to Python 2.6?
from Utils import zipfile

log = logging.getLogger('downloader')

retrycodes = urlgrabber.grabber.URLGrabberOptions().retrycodes
# 12 is the error number when a connection times out opening a socket.
if 12 not in retrycodes:
   retrycodes.append(12)

def interrupt_callback(cbobj):
   raise KeyboardInterrupt

# Set defaults for urlgrabber
urlgrab_opts = {'retrycodes':retrycodes,
                'timeout': 60.0,    # float # of seconds
                'retry':   10,
                'interrupt_callback': interrupt_callback,
                'throttle': None}

class DownloaderError(Exception):
   def __init__(self, url, local, err):
      Exception.__init__(self, url, local, err)
      self.url = url
      self.local = local
      self.err = err

class UnzipError(Exception):
   def __init__(self, zipfile, msg):
      Exception.__init__(self, zipfile, msg)
      self.zipfile = zipfile
      self.msg = msg

class Downloader(object):
   '''Simple class to download a URL to the local filesystem or open a
      file stream from a URL.  Can update a progress bar.
   '''
   def __init__(self, url, local='', reget=False, progress=None, **urlgrabopts):
      '''* url      - URL to retrieve.
         * local    - Where to save retrieved file.  Set to '' for streaming
                      mode.
         * reget    - Replace local file, even if it is same size as remote.
         * progress - If provided, should be on object which provides the
                      following methods:
                      * start()    - Called at the beginning of download with
                                     no arguments.
                      * set()      - Takes two arguments, the first is the
                                     amount completed and the second is the
                                     total.
                      * show()     - Takes no arguments, and causes the
                                     progress to be shown on screen.
                      * setTopic() - Takes one argument, the name of the file
                                     being retrieved.
                      * stop()     - Takes no arguments, called at finish.
         * urlgrabopts - Extra keyword arguments can be passed in for urlgrabber.
                         They only affect this instance of Downloader and not
                         the global urlgrab_opts of this module.
                         See urlgrabber.grabber documentation.
      '''
      # urlgrabber does jacked up things with file:/// URLs on Windows, so
      # let's just handle them as regular paths.
      scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
      if scheme == "file":
         self.url = urllib.url2pathname(path)
      else:
         self.url = url
      host, port = urllib.splitport(netloc)
      # deal with IPv6 url case
      host.strip("[]")
      self.netloc = "%s:%s" % (host, port)
      self.local = local
      self.reget = reget
      self.progress = None
      if progress is not None:
         self.progress = Downloader._progressAdapter(progress)
      self.urlgrabopts = urlgrab_opts
      self.urlgrabopts.update(urlgrabopts)

   class _progressAdapter(object):
      '''This class is to translate between urlgrabber''s progress API and the
         one implemented by SMART.'''
      def __init__(self, progress):
         self.progress = progress
         self.progress.start()

      def start(self, filename, url, basename, length, text):
         self.progress.setTopic(basename)
         if length is None:
            self.length = 0
         else:
            self.length = length

      def update(self, read):
         self.progress.set(read, self.length)
         self.progress.show()

      def end(self, bytes=0):
         self.progress.stop()

   def _getrealurl(self, **urlgrab_opts):
      # Translate datastore path to absolute file path
      if PathUtils.IsDatastorePath(self.url):
         self.url=PathUtils.DatastoreToFilepath(self.url)
      # We need to do a separate urlopen to avoid starting progress bars
      # that we might not finish.
      u = grabber.urlopen(self.url, **urlgrab_opts)
      # urllib has a bug where it does not raise an exception if an FTP file
      # doesn't exist.  In that case, it returns zero-length data.  We
      # catch that case here by look ing for zero-length data.  Note that we
      # don't explicitly check for FTP URLs, since anything that doesn't
      # have any data is surely invalid anyway.
      data = u.read(2)
      u.close()
      if not len(data):
         raise DownloaderError(self.url, self.local,
                                 'Zero-length or non-existent file.')
      # If the URL is redirected, override self.url to use the redirect.
      realurl = u.fo.geturl()
      if not realurl.startswith('file:') and realurl != self.url:
         log.debug('Using redirect URL %s' % realurl)
         self.url = realurl
      return u

   def _getfromurl(self):
      if os.path.exists(self.url):
         self.local = os.path.normpath(self.url)
         return self.local
      # We build our own urllib2.OpenerDirector and pass it to urlgrabber.
      # This is to work around some broken HTTPS proxy support in the modules
      # distributed with Python.
      # Note that passing in 'None' is OK.
      if os.name == 'nt' and urllib.proxy_bypass_registry(self.netloc):
         proxy_handler = Utils.Proxy.ProxyHandler({})
      else:
         proxy_handler = Utils.Proxy.ProxyHandler(self.urlgrabopts.get('proxies'))
      self.urlgrabopts['opener'] = urllib2.build_opener(proxy_handler)

      try:
         log.debug('Downloading %s to %s...' % (self.url, self.local))
         u = self._getrealurl(**self.urlgrabopts)
         if not self.reget:
            # Note:  urlgrabber fakes a 'content-length' header for file
            #        URLs, too, so this works even for local paths.
            if u.hdr.has_key('Content-Length') and os.path.exists(self.local):
               size = os.lstat(self.local)[stat.ST_SIZE]
               if str(size) == u.hdr['Content-Length']:
                  return self.local

         self.local = grabber.urlgrab(self.url, filename=self.local,
                                      progress_obj=self.progress,
                                      **self.urlgrabopts)

         return self.local
      except urlgrabber.grabber.URLGrabError, e:
         raise DownloaderError(self.url, self.local, str(e))

   def _openfromurl(self):
      if os.path.exists(self.url):
         # Default 'r' mode won't work for non-text files on Windows
         return open(self.url, 'rb')
      # We build our own urllib2.OpenerDirector and pass it to urlgrabber.
      # This is to work around some broken HTTPS proxy support in the modules
      # distributed with Python.
      if os.name == 'nt' and urllib.proxy_bypass_registry(self.netloc):
         proxy_handler = Utils.Proxy.ProxyHandler({})
      else:
         proxy_handler = Utils.Proxy.ProxyHandler(self.urlgrabopts.get('proxies'))
      self.urlgrabopts['opener'] = urllib2.build_opener(proxy_handler)

      try:
         log.debug('Downloading from %s...' % (self.url))
         self._getrealurl(**self.urlgrabopts)
         return grabber.urlopen(self.url, **self.urlgrabopts)
      except urlgrabber.grabber.URLGrabError, e:
         raise DownloaderError(self.url, self.local, str(e))

   @staticmethod
   def _openzipfile(zipfn, memberfn):
      z = zipfile.ZipFile(zipfn)
      namelist = z.namelist()
      if memberfn not in namelist:
         # Try to match ./path/to/meta, path/to/meta, /path/to/meta.
         normalized = memberfn.lstrip(os.sep + os.curdir)
         for name in namelist:
            if name.lstrip(os.sep + os.curdir) == normalized:
               memberfn = name
               break
      # Python 2.6 only
      return z.open(memberfn)

   def _openfromzip(self):
      try:
         zipfn, memberfn = self.url[4:].split('?', 1)
      except:
         msg = "Invalid Location: %s." % self.url
         raise DownloaderError(self.url, self.local, msg)

      try:
         return self._openzipfile(zipfn, memberfn)
      except Exception, e:
         msg = "Error extracting %s from %s: %s" % (memberfn, zipfn, e)
         raise DownloaderError(self.url, self.local, msg)

   def _getfromzip(self):
      try:
         zipfn, memberfn = self.url[4:].split('?', 1)
      except:
         msg = "Invalid Location: %s." % self.url
         raise DownloaderError(self.url, self.local, msg)

      try:
         src = self._openzipfile(zipfn, memberfn)
         dst = open(self.local, 'wb')
         shutil.copyfileobj(src, dst)
         src.close()
         dst.close()
         #not sure what to do about the line below - evan
         #z.close()
      except Exception, e:
         msg = "Error extracting %s from %s: %s" % (memberfn, zipfn, e)
         raise DownloaderError(self.url, self.local, msg)
      return self.local

   def Get(self):
      """Retrieves file, setting the object's 'local' attribute to the location
         of the file on the local file system. May also change the value of
         the object's 'url' attribute, if the HTTP server returns a redirect.
            Returns: absolute path to the local file. Note that the return
                     value may be different than the 'local' parameter passed
                     to the constructor. When the 'url' parameter is a local
                     path or file:// URL, the location of the file is returned
                     without making a copy. This conserves space on the file
                     system.
            Raises:
               * DownloaderError on failure.
      """
      assert self.local, "self.local cannot be empty"
      destdir = os.path.dirname(self.local)
      if not os.path.isdir(destdir):
         try:
            os.makedirs(destdir)
         except OSError, e:
            msg = "Cannot create directory %s: %s" % (destdir, e)
            raise Errors.FileIOError(destdir, msg)

      if self.url.startswith('zip:'):
         return self._getfromzip()
      else:
         return self._getfromurl()

   def Open(self):
      """Opens the url as a file object.  May change the value of the object's
         'url' attribute, if the HTTP server returns a redirect.
            Returns: a file object from which the callee can read directly
                     from the url to implement a download.
            Raises:
               * DownloaderError on failure.
      """
      if self.url.startswith('zip:'):
         return self._openfromzip()
      else:
         return self._openfromurl()


class CachedZipDownloader(Downloader):
   """
   Class that can cache, download, and unpack .zip files.
   Useful attributes:
     destdir  - download and unpack .zip in here
     local    - full local path of .zip file
   """
   def __init__(self, url, cachedir='/tmp', destdir=None, prefix="zip",
                destzipname=None, cleanupzip=False,
                reget=False, progress=None):
      """
       * url      - http:, file:, ftp:, or /path/to/zipfile.zip
       * cachedir - by default, download .zip to and unpack .zip in
                    <cachedir>/<prefix><hash(url)>/
       * prefix   - see above
       * destdir  - download and unpack .zip in <destdir> instead
       * destzipname - rename .zip after downloading
       * cleanupzip - delete .zip after unpacking to save space.
                    NB: if True, reget will have no effect.
      """
      if destdir:
         self.destdir = destdir
      else:
         self.destdir = os.path.join(cachedir,
                                     prefix + str(hash(url)))
      log.debug('Using %s for zip %s' % (self.destdir, url))
      if not destzipname:
         destzipname = os.path.basename(url)
      local = os.path.join(self.destdir, destzipname)

      Downloader.__init__(self, url, local, reget, progress)
      self.cleanupzip = cleanupzip

   def GetAndUnpack(self):
      """
      Downloads and unpacks .zip file.
      To save space, the .zip file is not downloaded if
      url starts with / (absolute path).
      If reget is false, zip is only redownloaded if
      .zip file differs in size from downloaded one.
      If cleanupzip is True, zip is deleted after
      unpacking, again to save space, if .zip was
      downloaded.

      Returns:
       path of downloaded .zip file

      Throws:
       DownloaderError - during actual download of .zip file
       UnzipError      - .zip file format error or corruption
       FileIOError     - cannot create destdir
      """
      if not os.path.isdir(self.destdir):
         #
         # Normally OSErrors are caught in esxupdate, but
         # the error string for os.makedirs is terribly confusing,
         # so we intervene here.
         #
         try:
            os.makedirs(self.destdir)
         except OSError, e:
            raise Errors.FileIOError(self.destdir, 'Cannot create dir %s: %s'
                              % (self.destdir, str(e)))

      local = self.Get()

      #
      # zipfile has a bug, returning IOError if .zip size is less than
      # 22 bytes, due to a seek(-22,2).  There is no I/O error here.
      # Report this as UnzipError - smallest zip file I could create
      # using 1 char file was 138 bytes.
      if os.path.getsize(local) < 100:
        raise UnzipError(local, "File %s is too small to be a .zip file"
                         % (local))

      try:
         z = zipfile.ZipFile(local)
         for info in z.filelist:
            # 0x10 is MS-DOS attribute bit for a directory.  If we don't check
            # for this, the zipfile module creates an empty file for the entry,
            # which causes problems when other archive members have it in their
            # path as a directory.  See PR 338908.
            if info.external_attr & 0x10: continue
            z.extract(info, path=self.destdir)
         z.close()
      except zipfile.error, e:
         raise UnzipError(local, "Error unzipping %s: %s" %
                          (local, str(e)) )

      if self.cleanupzip and local == self.local:
         os.unlink(local)

      return local


def SetProxy(urlport=None):
   """
   Sets HTTP/FTP/HTTPS proxy for urlgrabber to urlport, of 'url:port'
   format, such as 'http://proxy.abc.com:3128'
   If urlport is None or no argument given, the proxies option is
   deleted from urlgrab_opts.
   """
   if urlport:
     urlgrab_opts['proxies'] = {'http': urlport,
                                'https': urlport,
                                'ftp': urlport}
   elif 'proxies' in urlgrab_opts:
     del urlgrab_opts['proxies']

def SetTimeout(timeout=None):
   '''
   Sets downloader timeout value
   '''
   if timeout:
      urlgrab_opts['timeout'] = timeout
   elif 'timeout' in urlgrab_opts:
      del urlgrab_opts['timeout']

def SetRetry(retry=None):
   '''Set number of times to attempt retry.  Set to 0 for infinite retries.
   '''
   if retry:
      urlgrab_opts['retry'] = retry
   elif 'retry' in urlgrab_opts:
      del urlgrab_opts['retry']

def SetRateLimit(limit=None):
   """Set the maximum rate, in bytes per second, that the downloader should
      read from a remote sources. The value may either be None to indicate no
      limit, or a positive integer.
   """
   urlgrab_opts["throttle"] = limit is None and limit or int(limit)
