#!/usr/bin/python
########################################################################
# Copyright (C) 2010 VMWare, Inc.                                      #
# All Rights Reserved                                                  #
########################################################################

import logging
import os
import shutil
from socket import gethostname

from vmware import runcommand

from vmware.esximage import Database
from vmware.esximage import Errors
from vmware.esximage import ImageProfile
from vmware.esximage import Vib
from vmware.esximage.Installer import Installer

BACKUP_SH = '/sbin/backup.sh'

class InvalidBootcfgError(Exception):
   pass

class InvalidBootbankError(Exception):
   pass

log = logging.getLogger(os.path.basename(__file__))

class Bootcfg(object):
   """Bootcfg encapsulates the boot.cfg file of a bootbank.

      Class Variables:
         * BOOTSTATE_SUCCESS   - The bootbank has booted successfully.
         * BOOTSTATE_UPDATED   - The bootbank is upgraded, but not booted yet.
         * BOOTSTATE_ATTEMPTED - The bootbank is being booted, but not finish
                                 yet.
         * BOOTSTATE_EMPTY     - The bootbank is empty.
         * BOOTSTATE_STAGED    - New IP has been staged to bootbank, but update
                                 is not finished yet.

      Attributes:
         * kernel    - Kernel name
         * kernelopt - The options to kernel module
         * modules   - List of modules to load. The list is constructed and
                       ordered by ImageProfile class.
         * build     - Version info of core ESX module, usually in
                       'version-build' format.
         * updated   - The updated sequential number.
         * bootstate - Bootbank state, one of BOOTSTATE_TYPES.
   """
   BOOTSTATE_SUCCESS   = 0
   BOOTSTATE_UPDATED   = 1
   BOOTSTATE_ATTEMPTED = 2
   BOOTSTATE_EMPTY     = 3
   BOOTSTATE_STAGED    = 4
   BOOTSTATE_TYPES = (BOOTSTATE_SUCCESS, BOOTSTATE_UPDATED,
         BOOTSTATE_ATTEMPTED, BOOTSTATE_EMPTY, BOOTSTATE_STAGED)
   INSTALLER_BOOTSTATES = (BOOTSTATE_EMPTY, BOOTSTATE_STAGED, BOOTSTATE_UPDATED)

   def __init__(self, path):
      """
         Parameters:
            * path - File path to boot.cfg.
      """
      self._path = path
      self.Clear()

   path = property(lambda self: self._path)

   def Clear(self):
      """Mark the bootbank as empty."""
      self.kernel = ''
      self.kernelopt = ''
      self.modules = []
      self.build = ''
      self.updated = -1
      self.bootstate = self.BOOTSTATE_EMPTY

   def Load(self):
      """Load and parse boot.cfg file.
         Exceptions:
            * InvalidBootcfgError - If not all of the expected keys are loaded,
                                   an error in parsing the value, or an error to
                                   open the file.
      """
      self.Clear()
      expectedkeys = ['updated', 'bootstate', 'kernel', 'kernelopt', 'modules',
            'build']
      try:
         for line in open(self._path):
            line = line.strip()
            if line:
               k, sep, v = line.partition('=')
               if sep != '=':
                  continue

               k = k.strip()
               v = v.strip()
               if k == 'updated':
                  self.updated = int(v)
               elif k == 'bootstate':
                  self.bootstate = int(v)
               elif k == 'kernel':
                  self.kernel = v
               elif k == 'kernelopt':
                  self.kernelopt = v
               elif k == 'modules':
                  self.modules = v.split(' --- ')
               elif k == 'build':
                  self.build = v
               else:
                  log.info('Unrecognized value "%s" in boot.cfg' % (line))

               if k in expectedkeys:
                  expectedkeys.remove(k)
      except Exception, e:
         self.Clear()
         msg = 'Error parsing bootbank boot.cfg file %s: %s' % (self.path, e)
         raise InvalidBootcfgError(msg)

      if expectedkeys:
         self.Clear()
         msg = '%s value is expected in boot.cfg %s, but not found' % (
               expectedkeys, self._path)
         raise InvalidBootcfgError(msg)

   def Save(self):
      """Save Bootcfg instance content to file.
         Exceptions:
            * InstallationError - If there is an error in saving Bootcfg to file.
      """
      try:
         fp = open(self.path + '.tmp', 'w')
         c = ['bootstate=%s' % (str(self.bootstate)),
            'kernel=%s' % (self.kernel),
            'kernelopt=%s' % (self.kernelopt),
            'modules=%s'% (' --- '.join(self.modules)),
            'build=%s' % (self.build),
            'updated=%s' % (str(self.updated))]
         fp.write('\n'.join(c))
         fp.write('\n')
         fp.close()
         os.rename(self.path + '.tmp', self.path)
      except EnvironmentError, e:
         msg = 'Failed to save Bootcfg to file %s: %s' % (self._path, e)
         raise Errors.InstallationError(None, msg)

   def UpdateState(self, state):
      """Change bootstate value to BOOTSTATE_EMPTY, BOOTSTATE_STAGED or
         BOOTSTATE_UPDATED, which are managed by Installer.

            Parameters:
               * state - one of BOOTSTATE_EMPTY, BOOTSTATE_STAGED,
                         BOOTSTATE_UPDATED
            Exceptions:
               * ValueError - If the given state is not managed by Installer.

      """
      if state in self.INSTALLER_BOOTSTATES:
         self.bootstate = state
      else:
         msg = 'Can not change state to %s, only %s are allowed' % (
               state, self.INSTALLER_BOOTSTATES)
         raise ValueError(msg)

class BootBank(object):
   """Encapsulates attributes of a single bootbank.

      Class Variables:
         * DB_FILE  - File name of bootbank package database.
         * BOOT_CFG - File name of the config file for bootloader

      Attributes:
         * db      - An instance of Database.TarDatabase representing the
                     bootbank package database.
         * updated - "updated" serial number from boot.cfg.  If the bootbank
                     contains an invalid image or boot.cfg, this should be -1.
         * path    - The full file path for the mounted bootbank.
   """
   DB_FILE = 'imgdb.tgz'
   BOOT_CFG = 'boot.cfg'

   def __init__(self, bootbankpath):
      """
         Exceptions:
            InvalidBootbankError - thrown if a valid bootbank could not be found
                              at the bootbankpath
      """
      if not os.path.isdir(bootbankpath):
         raise InvalidBootbankError("%s is not a directory!" % (bootbankpath))

      self._path = bootbankpath
      dbpath = os.path.join(bootbankpath, self.DB_FILE)
      self.db = Database.TarDatabase(dbpath)
      bootcfgpath = os.path.join(bootbankpath, self.BOOT_CFG)
      self.bootcfg = Bootcfg(bootcfgpath)

   def __str__(self):
      return "<BootBank:%s>" % (self.path)

   @property
   def path(self):
      return self._path

   def _getupdated(self):
      return self.bootcfg.updated

   def _setupdated(self, updated):
      self.bootcfg.updated = updated

   updated = property(_getupdated, _setupdated)

   @property
   def bootstate(self):
      return self.bootcfg.bootstate

   def Load(self, raiseerror=True, createprofile=False):
      """Load database and boot.cfg from the bootbank
         Exceptions:
            * InvalidBootbankError - If raiseerror is True and there is an error
                                     loading boot.cfg or database.
      """
      try:
         self.bootcfg.Load()
      except InvalidBootcfgError, e:
         if raiseerror:
            msg = 'Error in loading boot.cfg from bootbank %s: %s' % (
                  self._path, e)
            raise InvalidBootbankError(msg)

      if self.bootcfg.bootstate in (Bootcfg.BOOTSTATE_UPDATED,
            Bootcfg.BOOTSTATE_STAGED, Bootcfg.BOOTSTATE_SUCCESS):
         try:
            self.db.Load()
         except (Errors.DatabaseFormatError, Errors.DatabaseIOError), e:
            self.db.Clear()
            self.bootcfg.UpdateState(Bootcfg.BOOTSTATE_EMPTY)
            if raiseerror:
               msg = 'Error in loading database for bootbank %s: %s' % (
                  self._path, e)
               raise InvalidBootbankError(msg)

         if self.db.profile is None and createprofile:
            log.debug('Creating an empty ImageProfile for bootbank %s' % (
               self._path))
            self.db.Clear()
            pname = 'Empty ImageProfile'
            creator = gethostname()
            p = ImageProfile.ImageProfile(pname, creator)
            self.db.profiles.AddProfile(p)

         if self.db.profile is not None:
            self.db.profile.vibs = self.db.vibs

   def Clear(self):
      """Clear bootbank DB to empty and flag the bootbank as empty
         Caller is responsible to call Save to make the change persistent.
      """
      self.bootcfg.UpdateState(self.bootcfg.BOOTSTATE_EMPTY)
      self.db.Clear()

   def Save(self):
      """Flush bootcfg and database to persistent storage.
      """
      self.db.Save()
      self.bootcfg.Save()

   def ClearContent(self):
      """Mark the bootbank as empty in database and boot.cfg. Files other than
         database and boot.cfg are removed as well.
      """
      try:
         self.Clear()
         self.Save()
      except Exception, e:
         msg = 'Error in clearing bootcfg or DB for bootbank %s: %s' % (
               self._path, e)
         raise Errors.InstallationError(None, msg)

      try:
         for root, dirs, files in os.walk(self.path, topdown=True):
            for dn in dirs:
                  shutil.rmtree(os.path.join(root, dn))
            for name in files:
               if name not in (BootBank.BOOT_CFG, BootBank.DB_FILE):
                  os.unlink(os.path.join(root, name))
      except EnvironmentError, e:
         msg = 'Failed to clear bootbank content %s: %s' % (self._path, e)
         raise Errors.InstallationError(None, msg)

   def Recover(self):
      """Check file system level error and try to repair.
      """
      from vmware.esximage.Utils import HostInfo
      rc = HostInfo.CheckAltbootbankFS()
      if rc != 0:
         msg = ('There was an error checking file system on altbootbank, '
                'please see log for detail.')
         raise Errors.InstallationError('', msg)

   def Verify(self):
      """ Verify whether the content of bootbank is consistent with imgprofile
      in DB.
      """
      #TODO: verify payloads according to DB

      #Verify vib payload

   def Stage(self, srcbank, keeps, imgprofile):
      """Stage the initial content of this bootbank. If srcbank is different
         from this bootbank, for VIBs in keeps list, payloads will be
         copied to this bootbank and renamed according to imgprofile. If
         srcbank is this bootbank, VIBs not in keeps list will be removed from
         the bootbank and the VIBs in keeps list will be renamed according to target
         image profile.

         Parameters:
            * srcbank    - BootBank instance, either 'bootbank' or
                           'altbootbank', the content is used to build target
                           image profile as source input.
            * keeps      - List of id of VIBs which need to be kept from srcbank
                           to construct target imageprofile.
            * imgprofile - Target image profile for this bootbank.

         Exceptions:
            * InstallationError - If there is an error to stage the imgprofile
                                  to this bootbank
      """
      if srcbank.path == self.path:
         self._inplaceStage(keeps, imgprofile)
      else:
         self._copyStage(srcbank, keeps, imgprofile)

      # Database is updated to the target ImageProfile, though the content is
      # not ready yet. It is fine because this bootbank is considered as
      # empty after calling stage.
      self.db.vibs.clear()
      self.db.vibs = imgprofile.vibs
      self.db.profiles.clear()
      self.db.profiles.AddProfile(imgprofile)

   def _inplaceStage(self, keeps, imgprofile):
      #NOTE: This only uses current bootbank VIB content, we might do better in some
      # case to use VIBs from /bootbank.
      self.bootcfg.UpdateState(self.bootcfg.BOOTSTATE_EMPTY)
      self.bootcfg.Save()
      kept = set()

      for vibid in self.db.profile.vibstates:
         # In place renaming to a temporary name
         vibstate = self.db.profile.vibstates[vibid]
         if vibid in keeps:
            deststate = imgprofile.vibstates[vibid]
            for pl in self.db.vibs[vibid].payloads:
               if pl.payloadtype not in BootBankInstaller.SUPPORTED_PAYLOADS:
                  continue

               plname = pl.name
               srcname = None
               if plname in vibstate.payloads:
                  srcname = vibstate.payloads[plname]

               if srcname is not None:
                  destname = deststate.payloads[plname] + '.stg'
                  dst = os.path.join(self._path, destname)
                  src = os.path.join(self._path, srcname)

                  try:
                     os.rename(src, dst)
                  except OSError, e:
                     msg = 'Failed to '
                     raise Errors.InstallationError(str(imgprofile.vibIDs),
                           msg)
               else:
                  msg = ('Can not locate source for payload %s of VIB %s '
                         ' from bootbank %s' % (plname, vibstate.id, self._path))
                  raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

               vibstate.payloads[plname] = destname
            kept.add(vibstate.id)
         else:
            # Remove unwanted vibs
            log.info('Removing VIB %s from %s' % (vibid, self._path))
            for plname in vibstate.payloads:
               localfile = os.path.join(self.path, vibstate.payloads[plname])
               if not os.path.isfile(localfile):
                  continue

               log.info('Removing payload %s at %s' % (plname, localfile))
               try:
                  os.unlink(localfile)
               except EnvironmentError, e:
                  msg = ('Can not remove payload %s in VIB %s: %s' %
                        (localfile, vibid, e))
                  raise Errors.InstallationError(msg)
            del self.db.vibs[vibid]

      if kept != set(keeps):
         left = set(keeps) - kept
         msg = 'VIBs %s are not found in bootbank %s' % (left, self._path)
         raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

      # re-establish local filename
      # In order to avoid name collision between original localname and
      # target localname, payload is renamed to a name space ends with
      # '.stg' and then move to the final name.
      for vibid in keeps:
         vibstate = self.db.profile.vibstates[vibid]
         for pl in self.db.vibs[vibid].payloads:
            if pl.payloadtype not in BootBankInstaller.SUPPORTED_PAYLOADS:
               continue

            plname = pl.name
            src = os.path.join(self.path, vibstate.payloads[plname])
            dst = src[:-4]
            if src.endswith('.stg'):
               try:
                  os.rename(src, dst)
               except OSError, e:
                  msg = ('Failed to rename temporary payload file %s '
                          'to final name %s: %s.' % (src, dst, e))
                  raise Errors.InstallationError(str(imgprofile.vibIDs), msg)
            else:
               msg = 'Unexpected temporary payload file name %s' % (src)
               raise Errors.InstallationError(str(imgprofile.vibIDs), msg)
            vibstate.payloads[plname] = vibstate.payloads[plname][:-4]

   def _copyStage(self, srcbank, keeps, imgprofile):
      # Preparation altbootbank based on srcbank
      # TODO: we can do better if there is a partial databse
      # Let's try the hard way first - just copy src from bootbank
      self.ClearContent()
      for vibid in keeps:
         if vibid not in srcbank.db.profile.vibIDs:
            msg = 'Can not find VIB %s in bootbank %s.' % (vibid,
                  srcbank.path)
            raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

         log.info('Copying payloads of %s from %s to %s' % (vibid,
            srcbank.path, self._path))
         deststate = imgprofile.vibstates[vibid]
         vibstate = srcbank.db.profile.vibstates[vibid]
         for pl in srcbank.db.vibs[vibid].payloads:
            if pl.payloadtype not in BootBankInstaller.SUPPORTED_PAYLOADS:
               continue

            plname = pl.name
            srcfile = None
            if plname in vibstate.payloads:
               srcfile = os.path.join(srcbank.path, vibstate.payloads[plname])

            if srcfile is not None:
               dstfile = os.path.join(self.path, deststate.payloads[plname])
               log.info('Copying %s to %s' % (srcfile, dstfile))
               try:
                  shutil.copy2(srcfile, dstfile)
               except EnvironmentError, e:
                  msg = 'Error in copying from %s to %s: %s' % (srcfile,
                        dstfile, e)
                  raise Errors.InstallationError(str(imgprofile.vibIDs), msg)
               #TODO: verify copy
               #log.debug('Verifying the file %s' % (dstfile))
            else:
               msg = ('Can not locate source for payload %s of VIB %s '
                      'from bootbank %s' % (plname, vibid, srcbank.path))
               raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

class BootBankInstaller(Installer):
   """BootBankInstaller is the Installer class for building
      a ESXi VFAT bootbank image.

      Class Variables:
         * BOOTBANK_STATE_FRESH   - altbootbank is either empty or a previous boot
                                    image.
         * BOOTBANK_STATE_STAGED  - altbootbank is staged with new ImageProfile,
                                    and VIBs has been installed to the bootbank,
                                    but the host will still boot from current
                                    bootbank in next reboot.
         * BOOTBANK_STATE_UPDATED - altbootbank is upgraded. The host will boot
                                    from altbootbank in next reboot.
         * SUPPORTED_PAYLOADS     - tuple of payload types that will be handled
                                    by the installer.

      Attributes:
         * bootbank    - A BootBank instance representing /bootbank, the bank
                         from which the host booted
         * altbootbank - A BootBank instance representing /altbootbank, the
                         bank holding the future (or previous) boot image
         * bootbankstate - Indicate whether altbootbank is staged or updated,
                           one of BOOTBANK_STATE_*.
   """
   installertype = "boot"
   priority = 10
   BOOTBANK = '/bootbank'
   ALTBOOTBANK = '/altbootbank'
   BUFFER_SIZE = 1024 * 1024

   # NOTE: 10MB for database, state.tgz
   BOOTBANK_PADDING_MB = 10

   BOOTBANK_STATE_FRESH = 0
   BOOTBANK_STATE_STAGED = 1
   BOOTBANK_STATE_UPDATED = 2

   SUPPORTED_VIBS = set([Vib.BaseVib.TYPE_BOOTBANK,])
   SUPPORTED_PAYLOADS = (Vib.Payload.TYPE_TGZ, Vib.Payload.TYPE_VGZ,
         Vib.Payload.TYPE_BOOT)

   def __init__(self, bootbank='/bootbank', altbootbank='/altbootbank'):
      """Constructor for this Installer class.
         /bootbank and /altbootbank must both exist. /bootbank must points to
         valid bootbank with boot.cfg file.  If not, this class is not
         instantiated.

         Exceptions:
            InstallerNotAppropriate - Environment is not appropriate for
               this type of installer.
      """
      #TODO: add probe to host environment, raise error if bootbank is not
      # supported.
      try:
         self.bootbank = BootBank(bootbank)
      except InvalidBootbankError, e:
         msg = 'bootbank is invalid: %s' % (e)
         raise Errors.InstallerNotAppropriate(self.installertype, msg)

      try:
         self.bootbank.Load(createprofile = True)
         self.altbootbank = BootBank(altbootbank)
         # Altbootbank is not always available.
         self.altbootbank.Load(raiseerror = False)
      except InvalidBootbankError, e:
         msg = 'altbootbank is invalid: %s' % (e)
         raise Errors.InstallationError(None, msg)
      self.problems = list()

   @property
   def bootbankstate(self):
      if self.bootbank.updated < self.altbootbank.updated:
         if self.altbootbank.bootstate == Bootcfg.BOOTSTATE_UPDATED:
            return self.BOOTBANK_STATE_UPDATED
         elif self.altbootbank.bootstate == Bootcfg.BOOTSTATE_STAGED:
            return self.BOOTBANK_STATE_STAGED
      return self.BOOTBANK_STATE_FRESH

   @property
   def database(self):
      if self.bootbankstate == self.BOOTBANK_STATE_UPDATED:
         return self.altbootbank.db
      else:
         return self.bootbank.db

   @property
   def stagedatabase(self):
      if self.bootbankstate == self.BOOTBANK_STATE_STAGED:
         return self.altbootbank.db
      else:
         return None

   def CheckInstallationSize(self, imgprofile):
      '''Calculate the installation size of imgprofile on bootbank.
         Verify bootbank can hold the image. BOOTBANK_PADDING_MB is reserved for
         state.tgz and other extra files.

         NOTE: No attempt is made to estimate the size of state.tgz, transaction
               might still fail if state.tgz needs more than the reserved padding
               space.

         Parameters:
            * imgprofile - Target image profile

         Exceptions:
            * InstallationError - Installation size of this image profile can't
                                  fit in bootbank partition.
      '''
      totalsize = 0
      for vibid in imgprofile.vibIDs:
         vib = imgprofile.vibs[vibid]
         if vib.vibtype in self.SUPPORTED_VIBS:
            for payload in vib.payloads:
               if payload.payloadtype in self.SUPPORTED_PAYLOADS:
                  totalsize += payload.size

      stat = os.statvfs(self.altbootbank.path)
      capacity = (stat.f_bsize * stat.f_blocks -
                  self.BOOTBANK_PADDING_MB * 1024 *1024)
      if totalsize > capacity:
         msg = ('The pending transaction requires %d MB free space, however '
                'the maximum supported size is %d MB.' % (totalsize/1024/1024,
                   capacity/1024/1024))
         log.error(msg)
         raise Errors.InstallationError('', msg)

   def CheckBootState(self):
      '''Check the bootstate of current bootbank. If bootstate is not
         BOOTSTATE_SUCCESS, raise InstallationError exception.
      '''
      if (self.bootbank.bootstate != Bootcfg.BOOTSTATE_SUCCESS):
         msg = ('Current bootbank %s is not verified and most likely a '
                'serious problem was encountered during boot, it is not safe '
                'to continue altbootbank install. bootstate is %d, expected '
                'value is %s.' % ( self.bootbank.path, self.bootbank.bootstate,
                 Bootcfg.BOOTSTATE_SUCCESS))
         raise Errors.InstallationError('', msg)

   def PreInstCheck(self, imgprofile):
      '''Perform sanity check for new image profile change.
         Parameters:
            * imgprofile - Target image profile

         Exceptions:
            * InstallationError - /bootbank is not in BOOTSTATE_SUCCESS state or
                                  installation size can't fit in bootbank
                                  partition.
      '''
      self.CheckBootState()
      self.CheckInstallationSize(imgprofile)

   def StartTransaction(self, imgprofile, preparedest=True, **kwargs):
      """Initiates a new installation transaction. Calculate what actions
         need to be taken.  Prepare /altbootbank by copying all contents
         from /bootbank.
         This method may change the installation destination.

         Parameters:
            * imgprofile - The ImageProfile instance representing the
                           target set of VIBs for the new image
            * preparedest - Boolean, if True, then prepare the destination.
                           Set to false for a "dry run", to avoid changing
                           the destination.
            * forcebootbank - Boolean, if True, skip install of live image
                           even if its eligible for live install
         Returns:
            A tuple (installs, removes, staged), installs and removes are list
            of VIB IDs for HostImage.Stage() to install to the destination and
            to remove from the destination, in order to make it compliant
            with imgprofile. If altbootbank has already staged the imgprofile,
            staged is True.
            If there is nothing to do, (None, None, False) is returned.
         Exceptions:
            InstallationError
      """
      try:
         self.altbootbank.Load(raiseerror=False)
      except:
         self.altbootbank.Clear()

      imgprofile = imgprofile.Copy()
      unsupported = imgprofile.vibIDs - self.GetSupportedVibs(imgprofile.vibs)
      for vibid in unsupported:
         imgprofile.RemoveVib(vibid)

      staged = False
      if (self.altbootbank.bootstate == Bootcfg.BOOTSTATE_STAGED):
         if self.altbootbank.db.profile is not None and \
            self.altbootbank.db.profile.vibIDs == imgprofile.vibIDs:
            staged = True

      adds, removes = imgprofile.Diff(self.database.profile)

      if staged:
         return (adds, removes, staged)

      if preparedest and (removes or adds):
         if self.bootbankstate == self.BOOTBANK_STATE_UPDATED:
            srcbootbank = self.altbootbank
         else:
            srcbootbank = self.bootbank
            self.altbootbank.Recover()

         keeps = set(imgprofile.vibIDs) - set(adds)
         self.altbootbank.Stage(srcbootbank, keeps, imgprofile)

      return (adds, removes, staged)

   def OpenPayloadFile(self, vibid, payload, read = True, write = False):
      """Creates and returns a File-like object for either reading from
         or writing to a given payload.  One of read or write must be True, but
         ready and write cannot both be true.

         Parameters:
            * vibid   - The Vib id containing the payload
            * payload - The Vib.Payload instance to read or write
            * read    - Set to True to get a File object for reading
                        from the payload.
            * write   - Set to True to get a File object for writing
                        to the payload.
         Returns:
            A File-like object, must support read (for read), write (for
            write), close methods.
            None if the desired read/write is not supported.
         Exceptions:
            AssertionError    - neither read nor write is True, or both are true
            InstallationError - Cannot open file to write
      """
      Installer.OpenPayloadFile(self, vibid, payload, read = read, write = write)
      if read == True:
         localfile = None
         if vibid in self.bootbank.db.profile.vibIDs:
            vibstate = self.bootbank.db.profile.vibstates[vibid]
            rootpath = self.bootbank.path
         elif self.bootbankstate in (self.BOOTBANK_STATE_UPDATED,
               self.BOOTBANK_STATE_STAGED) and \
               vibid in self.altbootbank.db.profile.vibIDs:
            vibstate = self.altbootbank.db.profile.vibstates[vibid]
            rootpath = self.altbootbank.path
         else:
            msg = 'Cannot find VIB %s in BootbankInstaller'
            raise Errors.InstallationError(None, msg)

         if payload.name in vibstate.payloads:
            localfile = os.path.join(rootpath, vibstate.payloads[payload.name])
         else:
            msg = ("VIB %s found in BootbankInstaller (%s), but payload %s is "
                   "not found in the VIB" % (vibid, rootpath, payload.name))
            raise Errors.InstallationError(msg)

         try:
            return open(localfile, 'r')
         except EnvironmentError, e:
            msg = 'Error in opening payload %s of VIB %s at %s for read: %s' % (
                  payload.name, vibid, localfile, e)
            raise Errors.InstallationError(None, msg)

      elif write == True:
         # bootbank installer only supports BOOT, VGZ and TGZ payload
         if payload.payloadtype not in self.SUPPORTED_PAYLOADS:
            log.debug("Payload %s of type '%s' in VIB '%s' is not supported by "
                      "BootbankInstaller." % (payload.payloadtype, payload.name,
                         vibid))
            return None

         bootbankfile = os.path.join(self.altbootbank.path, payload.localname)
         try:
            fp = open(bootbankfile, 'w')
         except EnvironmentError, e:
            msg = 'Can not open %s to write payload %s: %s' % (bootbankfile,
                  payload.name, e)
            raise Errors.InstallationError(None, msg)
         return fp

      return None

   def Cleanup(self):
      """Cleans up after a Transaction that has been started, perhaps from
         a previous instantiation of the Installer class.
         For /altbootbank, there is no real cleanup we can do.
      """
      try:
         self.altbootbank.bootcfg.UpdateState(Bootcfg.BOOTSTATE_EMPTY)
         self.altbootbank.bootcfg.Save()
      except Exception, e:
         pass

   def CompleteStage(self):
      """Complete the staging of /altbootbank by writing out the new imgdb
         and the new boot.cfg.

         Exceptions:
            InstallationError
      """
      imgprofile = self.altbootbank.db.profile
      bootorder = imgprofile.GetBootOrder(self.SUPPORTED_PAYLOADS,
                                          self.SUPPORTED_VIBS)

      #TODO: need to accomodate custom boot option bewteen kernel and regular
      # module
      self.altbootbank.bootcfg.build = imgprofile.GetEsxVersion()


      try:
         _, p = bootorder[0]
         self.altbootbank.bootcfg.kernel = p.localname
         modules = [p.localname for (vibid, p) in bootorder[1:]]
         self.altbootbank.bootcfg.modules = modules

         self.altbootbank.bootcfg.UpdateState(Bootcfg.BOOTSTATE_STAGED)
         self.altbootbank.updated = self.bootbank.updated + 1
         self.altbootbank.Save()
      except Exception, e:
         msg = 'Failed to finish bootbank stage: %s'  % (e)
         raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

   def CacheNewImage(self, srcinstlr):
      log.info('Caching LiveImage to altbootbank...')
      self.altbootbank.Recover()
      self.altbootbank.ClearContent()
      self.altbootbank.bootcfg.UpdateState(Bootcfg.BOOTSTATE_EMPTY)
      self.altbootbank.bootcfg.Save()

      profile = srcinstlr.database.profile.Copy()
      self.altbootbank.db.vibs = profile.vibs
      self.altbootbank.db.profiles.AddProfile(profile)

      # Read payloads from live image
      kernelpayloads = []
      for vibid in profile.vibIDs:
         for payload in srcinstlr.database.vibs[vibid].payloads:

            payload.localname = profile.vibstates[vibid].payloads[payload.name]

            # Get dest file obj, skip if payload is not supported
            dst = self.OpenPayloadFile(vibid, payload, write=True, read=False)
            if dst is None:
               continue

            log.debug("About to write payload '%s' of VIB %s to '%s'" % (payload.name,
               vibid, self.altbootbank.path))

            src = None
            # Get src from bootbank
            if vibid in self.bootbank.db.profile.vibstates \
                  and  (payload.name in
                        self.bootbank.db.profile.vibstates[vibid].payloads):
               vibstate = self.bootbank.db.profile.vibstates[vibid]
               srcfile = os.path.join(self.bootbank.path,
                     vibstate.payloads[payload.name])
               log.debug('Using %s from bootbank as source' % (srcfile))
               src = open(srcfile, 'r')
            # Get src from live image for newly installed module
            elif payload.payloadtype != payload.TYPE_BOOT:
               log.debug('Using source from %s' % (
                  srcinstlr.__class__.__name__))
               src = srcinstlr.OpenPayloadFile(vibid, payload, read=True,
                     write=False)
            else:
               log.error('Could not found payload %s from sources' % (
                  payload.name))

            if src == None:
               msg = ('Error in opening payload %s of VIB %s from installer '
                      '%s' % (payload.name, vibid, srcinstlr.installertype))
               raise Errors.InstallationError([vibid,], msg)

            filesize = 0
            while True:
               inbytes = src.read(self.BUFFER_SIZE)
               if not inbytes:
                  break
               dst.write(inbytes)
               filesize += len(inbytes)
            log.debug('Total of %d bytes were written.' % (filesize))
            src.close()
            dst.close()

      self.CompleteStage()
      return self.Remediate()

   def Remediate(self, checkmaintmode=False):
      """For bootbank installs, the remediation consists of a reboot. Since
         the core engine does not carry out the reboot, we simply NOP.
         Returns:
            A Boolean, True if a reboot is needed.
         Exceptions:
            HostNotChanged    - If no staged ImageProfile
            InstallationError -
      """
      self.altbootbank.Load(raiseerror=False)

      if self.bootbankstate != self.BOOTBANK_STATE_STAGED:
         msg = "Bootbank is not yet staged, nothing to remediate."
         raise Errors.HostNotChanged(msg)

      imgprofile = self.altbootbank.db.profile

      # TODO: Verify bootbank with respect to imgprofile
      self.altbootbank.Verify()

      self.altbootbank.bootcfg.modules.append(BootBank.DB_FILE)

      if os.path.isfile(BACKUP_SH):
         try:
            rc, out = runcommand.runcommand('%s 0 %s' % (BACKUP_SH,
               self.altbootbank.path))
         except runcommand.RunCommandError, e:
            msg = ('Failed to run %s: %s' % (BACKUP_SH, e))
            raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

         if rc != 0:
            msg = ('Error in running backup script: non-zero code returned.'
                   '\nreturn code: %d\noutput: %s' % (rc, out))
            raise Errors.InstallationError(str(imgprofile.vibIDs), msg)
      else:
         log.debug("Skipping backup script '%s', which was not found" % (BACKUP_SH))

      statefile = os.path.join(self.altbootbank.path, 'state.tgz')
      if os.path.isfile(statefile):
         self.altbootbank.bootcfg.modules.append('state.tgz')

      log.info("boot config of '%s' is being updated, %d" %
            (self.altbootbank.path, self.altbootbank.updated))
      self.altbootbank.updated = self.bootbank.updated + 1
      self.altbootbank.bootcfg.UpdateState(Bootcfg.BOOTSTATE_UPDATED)

      # Keep kernel option from current bootbank a customer might have specified
      # The boot options stored in esx.conf are persisted by backup.sh script in
      # useropts.gz
      self.altbootbank.bootcfg.kernelopt = self.bootbank.bootcfg.kernelopt

      try:
         self.altbootbank.bootcfg.Save()
      except Exception, e:
         msg = 'Failed to remediate the host: %s' % (e)
         raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

      return True
