#!/usr/bin/python
########################################################################
# Copyright (C) 2010 VMWare, Inc.                                      #
# All Rights Reserved                                                  #
########################################################################

import fnmatch
import logging
import os
import re
import shutil

from vmware import runcommand

from vmware.esximage import Database
from vmware.esximage import Errors
from vmware.esximage import Vib
from vmware.esximage import VibCollection
from vmware.esximage.ImageProfile import ImageProfile
from vmware.esximage.Installer import Installer
from vmware.esximage.Utils import EsxGzip
from vmware.esximage.Utils import HostInfo
from vmware.esximage.Utils import PathUtils

log = logging.getLogger('LiveImageInstaller')
LIVEINST_MARKER = 'var/run/update_altbootbank'
SFCBD_RESTART_MARKER = 'var/run/sfcbRestartRequested'

FAILURE_WARNING = '''It is not safe to continue. Please reboot the host \
immediately to discard the unfinished update.'''

#
# If an image is live staged, we don't need to stage it to bootbank again
#
class LiveImage(object):
   '''Encapsulates attributes of LiveImage
      Attributes:
         * root             - root of the LiveImage, default '/'
         * stageroot        - root of the staging directory
         * stagedatadir     - Temporary directory for payloads files (unziped)
         * database         - An instance of Database for live image
         * stagedatabase    - An instance of Database.Database representing the
                              package database for staged image. None if not
                              yet staged.
         * isstaged         - Boolean, indicate whether there is a staged image
   '''
   DB_DIR = os.path.join('var', 'db', 'esximg')
   STAGE_DIR = 'tmp/img-stg'
   CHKCONFIG_DB = "etc/chkconfig.db"
   SERVICES_DIR = "etc/init.d"

   def __init__(self, root = '/', stagedir = None):
      self._root = root
      self.database = Database.Database(os.path.join(self._root, self.DB_DIR),
            dbcreate = True)
      if stagedir is None:
         stagedir = self.STAGE_DIR
      self._stageroot = os.path.join(self._root, stagedir)
      self._stagedata = os.path.join(self._stageroot, 'data')
      self._stagedbpath = os.path.join(self._stageroot, 'imgdb')
      self._stageindicator = os.path.join(self._stageroot, 'staged')
      self._stagedb = None
      self._tardisksdir = os.path.join(root, 'tardisks')
      self._chkconfigdb = os.path.join(root, self.CHKCONFIG_DB)
      self._servicesdir = os.path.join(root, self.SERVICES_DIR)
      self._servicestates = {"enabled": set(), "disabled": set(),
                             "added": set(), "removed": set(),
                             "upgradevibs": set()}
      self._triggers = [CIMOMRestartTrigger(),
                        WSMANRestartTrigger(),
                        FirewallRefreshTrigger(),
                        HostServiceRefreshTrigger(),]

      self.Load()

   @property
   def root(self):
      return self._root

   @property
   def stageroot(self):
      return self._stageroot

   @property
   def stagedatadir(self):
      return self._stagedata

   @property
   def stagedatabase(self):
      if self.isstaged:
         return self._stagedb
      return None
   @property
   def tardisksdir(self):
      return self._tardisksdir

   @property
   def chkconfigdb(self):
      return self._chkconfigdb

   @property
   def isstaged(self):
      return os.path.isfile(self._stageindicator)

   def Load(self):
      '''Load LiveImage and staging info.

         Exceptions:
            * DatabaseFormatError
            * DatabaseIOError
      '''
      self.database.Load()
      if self.isstaged:
         try:
            self._stagedb = Database.Database(self._stagedbpath, dbcreate=False)
            self._stagedb.Load()
         except Exception, e:
            msg = ('Found a copy of staged image, but unable to load the database: '
                   '%s. Resetting live image as non-staged' % (e))
            log.warn(msg)
            self._UnSetStaged()
      # profile.vibs should contain Vib instance as Transaction only replies on
      # image profile to get detailed VIB info.
      if self.database.profile is not None:
         self.database.profile.vibs = self.database.vibs
      if self.stagedatabase and self.stagedatabase.profile is not None:
         self.stagedatabase.profile.vibs = self.stagedatabase.vibs

   def StartTransaction(self, imgprofile):
      '''Reset staging directory to empty.
      '''
      problems = self._CheckHost(imgprofile)
      if problems:
         msg = 'Not enough resources to run the transaction: %s' % (problems)
         raise Errors.InstallationError([], msg)

      self._Setup()
      self._stagedb = Database.Database(self._stagedbpath, addprofile=False)
      self._stagedb.vibs = imgprofile.vibs
      self._stagedb.profiles.AddProfile(imgprofile)

   def CompleteStage(self):
      '''Save staging database and staged indicator
      '''
      self._stagedb.Save()
      try:
         self._SetStaged()
      except Exception, e:
         msg = 'Failed to change the status of the staged image: %s' % (e)
         raise Errors.InstallationError([], msg)

   def _MaintenanceModeVibs(self, tgtvibs, adds, removes,
         checkmaintmode=True):
      '''Create list of vibs which require maintenance mode to finish the
         transaction.
         Parameters:
            * tgtvibs - VibCollection instance for the desired image profile
                        vibs
            * adds    - List of Vib IDs which will be added during the
                        transaction
            * removes - List of Vib IDs which will be removed the during the
                        transaction
         Returns: A tuple (mmoderemove, mmodeinstall); mmoderemove
            is a set of VIB IDs which requires maintenance mode to be removed;
            mmodeinstall is a set of VIB IDs which requires maintenance mode to
            be installed.
      '''
      mmoderemove = set()
      mmodeinstall = set()
      for vibid in adds:
         if tgtvibs[vibid].maintenancemode.install:
            mmodeinstall.update([vibid,])
      for vibid in removes:
         if self.database.vibs[vibid].maintenancemode.remove:
            mmoderemove.update([vibid,])
      return (mmoderemove, mmodeinstall)

   def Remediate(self, checkmaintmode=True):
      '''Live remove VIBs and enable new VIBs
      '''
      if self.stagedatabase.profile is None:
         msg = ('There is an error in the DB of staging directory: %s, Please '
                'try the update command again.' % (self._stagedbpath))
         raise Errors.InstallationError([], msg)

      adds, removes = self.stagedatabase.profile.Diff(self.database.profile)
      vibstoadd = [self.stagedatabase.vibs[vibid] for vibid in adds]
      vibstoremove = [self.database.vibs[vibid] for vibid in removes]

      self._UpdateServiceStates(vibstoadd, vibstoremove)
      self._UpdateTriggers(vibstoadd, vibstoremove)

      # check maintenance mode required for the transaction
      mmoderemove, mmodeinstall = self._MaintenanceModeVibs(
            self.stagedatabase.vibs, adds, removes)
      if mmoderemove or mmodeinstall:
         msg = ('MaintenanceMode is required to '
                'remove: [%s]; install: [%s].'
                   % (', '.join(mmoderemove),
                      ', '.join(mmodeinstall)))
         log.debug(msg)
         if checkmaintmode:
            mmode = HostInfo.GetMaintenanceMode()
            if not mmode:
               raise Errors.MaintenanceModeError(msg)
         else:
            log.warning('MaintenanceMode check was skipped...')

      # save config files
      cmd = '/sbin/backup.sh 0'
      try:
         RunCmdWithMsg(cmd, raiseexception=False)
      except runcommand.RunCommandError, e:
         log.warning('There was an error to run [%s]: %s' % (cmd, e))

      # XXX Verify staged image

      self._RemoveVibs(removes)
      self._AddVibs(adds)

      # Run secpolicytools, PR 581102
      # 1. setup policy rules for vmkaccess (/etc/vmware/secpolicy/*)
      # 2. Apply the security label for any new tardisks that have been
      #    installed and mounted.
      cmd = '/sbin/secpolicytools -p'
      try:
         RunCmdWithMsg(cmd, raiseexception=False)
      except runcommand.RunCommandError, e:
         log.warning('There was an error to run [%s]: %s' % (cmd, e))

      # Start daemons in order
      problem = None
      failedservices = self._StartServices()
      if failedservices:
         problem = 'The following services were not started: %s\n%s' % (
               ' ,'.join(failedservices), FAILURE_WARNING)

      # Invoke post inst triggers
      for trigger in self._triggers:
         trigger.Run()

      self._UpdateDB()
      if problem is not None:
         raise Errors.InstallationError([], problem)

      # mark system live installed
      markerfile = os.path.join(self.root, LIVEINST_MARKER)
      msg = ("There was an error creating flag file '%s',"
             "configuration changes after installation might be lost "
             "after reboot." % (markerfile))
      try:
         open(markerfile, 'w').close()
      except IOError, e:
         log.warning("Creating file failed: %s" % e)
         log.warning(msg)


   def Cleanup(self):
      '''Remove data in staging area.
      '''
      # this is a bit tricky, have to track init, tardisk mapping,
      try:
         if os.path.exists(self._stageroot):
            shutil.rmtree(self._stageroot)
      except EnvironmentError, e:
         msg = 'Cannot remove live image staging directory: %s' % (e)
         raise Errors.InstallationError([], msg)
      self._stagedb = None

   def _RemoveVibs(self, removes):
      log.debug('Starting to live remove VIBs: %s' % (', '.join(removes)))

      shutdownscripts = set()
      for vibid in removes:
         vib = self.database.vibs[vibid]
         log.info('Live removing %s-%s' % (vib.name, vib.version))
         # run shutdown script
         scripts = self._GetFilesFromVib(vib,
               lambda x: fnmatch.fnmatch(x, "etc/vmware/shutdown/shutdown.d/*"))
         shutdownscripts.update(scripts)

      self._ShutdownServices()

      # Module unloading
      log.debug('Starting to run etc/vmware/shutdown/shutdown.d/*')
      for script in shutdownscripts:
         RunCmdWithMsg(script)

      # umount tardisks
      for vibid in removes:
         for name, localname in \
               self.database.profile.vibstates[vibid].payloads.items():
            log.debug('Trying to unmount payload [%s] of VIB %s' % (name, vibid))
            self._UnmountTardisk(localname)
      log.debug('done')

   def _AddVibs(self, adds):
      log.debug('Starting to enable VIBs: %s' % (', '.join(adds)))

      jumpstartplugins = set()

      # ordering non-overlay and overlay VIB, so non-overlay VIBs being mounted
      # first
      regvibs = []
      overlayvibs = []
      for vibid in adds:
         vib = self.stagedatabase.vibs[vibid]
         if vib.overlay:
            overlayvibs.append(vibid)
         else:
            regvibs.append(vibid)

      for vibid in regvibs + overlayvibs:
         vib = self.stagedatabase.vibs[vibid]

         plugins = self._GetFilesFromVib(vib,
               lambda x: fnmatch.fnmatch(x, "usr/libexec/jumpstart/plugins/*"))
         for fn in plugins:
            jumpstartplugins.add(os.path.basename(fn))

         log.debug('Live installing %s-%s' % (vib.name, vib.version))

         for name, localname in \
            self.stagedatabase.profile.vibstates[vibid].payloads.items():
            log.debug('Trying to mount payload [%s]' % (name))
            self._MountTardisk(localname)

         #TODO: Resource group management

      if jumpstartplugins:
         self._StartJumpStartPlugins(jumpstartplugins)

      log.debug('Enabling VIBs is done.')

   def _UpdateDB(self):
      self.database.vibs = self.stagedatabase.vibs
      self.database.profiles = self.stagedatabase.profiles
      self.database.Save()
      self.Cleanup()

   def _Setup(self):
      self.Cleanup()
      try:
         os.makedirs(self._stageroot)
         os.makedirs(self._stagedata)
      except EnvironmentError, e:
         msg = 'There was a problem setting up staging area: %s' % (e)
         raise Errors.InstallationError([], msg)

   def _SetStaged(self):
      if not os.path.isfile(self._stageindicator):
         f = open(self._stageindicator, 'w')
         f.close()

   def _UnSetStaged(self):
      if os.path.exists(self._stageindicator):
         os.unlink(self._stageindicator)

   def _CheckHost(self, adds):
      '''TODO: Check the host environment to make sure:
            * enough memory to expand VIB
            * enough userworlds memory
      '''
      return []

   def _MountTardisk(self, tardisk):
      tardiskpath = os.path.join(self._stagedata, tardisk)
      cmd = 'mv %s /tardisks/' %(tardiskpath)
      RunCmdWithMsg(cmd, 'Mounting %s...' % (tardisk))

   def _UnmountTardisk(self, tardisk):
      cmd = 'rm /tardisks/%s' % (tardisk)
      RunCmdWithMsg(cmd, 'Unmounting %s...' % (tardisk))

   @staticmethod
   def _GetFilesFromVib(vib, matchfn):
      files = []
      for fn in vib.filelist:
         normfile = PathUtils.CustomNormPath(fn)
         if matchfn(normfile):
            # return script with absolute path so the script can be run from any
            # directory.
            files.append(os.path.join('/', normfile.strip()))
      return files

   @staticmethod
   def _GetUpgradeVibs(adds, removes):
      addvibs = VibCollection.VibCollection()
      for vib in adds:
         addvibs.AddVib(vib)
      removevibs = VibCollection.VibCollection()
      for vib in removes:
         removevibs.AddVib(vib)

      allvibs = VibCollection.VibCollection()
      allvibs += addvibs
      allvibs += removevibs

      upgradevibs = set()
      scanner = allvibs.Scan()
      for vibs in (addvibs, removevibs):
         upgradevibs.update(scanner.GetUpdatesSet(vibs))
         upgradevibs.update(scanner.GetDowngradesSet(vibs))

      return upgradevibs

   def _UpdateServiceStates(self, adds, removes):
      enabled = set()
      disabled = set()
      serviceadds = set()
      serviceremoves = set()

      # set of vibids involving upgrade/downgrade
      upgradevibs = self._GetUpgradeVibs(adds, removes)

      filterfn = lambda x: fnmatch.fnmatch(x, "etc/init.d/*")
      for vib in adds:
         files = self._GetFilesFromVib(vib, filterfn)
         serviceadds.update(files)
         if vib.id in upgradevibs:
            self._servicestates["upgradevibs"].update(files)
      for vib in removes:
         files = self._GetFilesFromVib(vib, filterfn)
         serviceremoves.update(files)
         if vib.id in upgradevibs:
            self._servicestates["upgradevibs"].update(files)

      upgrades = serviceadds & serviceremoves
      serviceadds -= upgrades
      serviceremoves -= upgrades

      if upgrades:
         cmd = ["/sbin/chkconfig", "-B", self._chkconfigdb, "-D",
               self._servicesdir, "-l"]
         rc, out = runcommand.runcommand(cmd)
         if rc:
            msg = "Failed to get chkconfig service list: %s" % out
            raise Errors.InstallationError([], msg)

         for line in out.splitlines():
            line = line.strip()
            if line.endswith(" on"):
               servicepath = os.path.join(self._servicesdir, line[:-3].strip())
               enabled.add(servicepath)
            elif line.endswith(" off"):
               servicepath = os.path.join(self._servicesdir, line[:-4].strip())
               disabled.add(servicepath)

         for script in upgrades:
            if script in enabled:
               self._servicestates["enabled"].add(script)
            elif script in disabled:
               self._servicestates["disabled"].add(script)
      self._servicestates["added"].update(serviceadds)
      self._servicestates["removed"].update(serviceremoves)

   def _GetEnabledServices(self):
      cmd = ["/sbin/chkconfig", "-B", self._chkconfigdb, "-D",
             self._servicesdir, "-i", "-o"]
      rc, out = runcommand.runcommand(cmd)
      if rc:
         msg = "Failed to obtain chkconfig service list: %s" % out
         raise Errors.InstallationError([], msg)
      return [line.strip() for line in out.splitlines()]

   def _ShutdownServices(self):
      '''Shutdown services according to the order in chkconfig.db. It is assumed
         that all the services are managed by chkconfig.
      '''
      tostop = self._servicestates["removed"] | self._servicestates["enabled"]
      for service in reversed(self._GetEnabledServices()):
         cmd = [service, 'stop']
         if service in self._servicestates["upgradevibs"]:
            cmd.append("upgrade")
         else:
            cmd.append("remove")
         if service in tostop:
            RunCmdWithMsg(cmd)

   def _SetServiceEnabled(self, service, enabled):
      service = os.path.basename(service)
      cmd = ["/sbin/chkconfig", "-B", self._chkconfigdb, "-D",
             self._servicesdir, service, (enabled and "on" or "off")]
      rc, out = runcommand.runcommand(cmd)
      if rc:
         action = enabled and "enable" or "disable"
         msg = "Unable to %s service %s\n%s" % (action, service,
               FAILURE_WARNING)
         raise Errors.InstallationError([], msg)

   def _StartServices(self):
      # Need to enable/disable any services which were previously explicitly
      # enabled/disabled.
      enabled = set(self._GetEnabledServices())
      for service in enabled & self._servicestates["disabled"]:
         self._SetServiceEnabled(service, False)

      for service in self._servicestates["enabled"] - enabled:
         self._SetServiceEnabled(service, True)

      tostart = self._servicestates["enabled"] | self._servicestates["added"]
      failedservices = list()
      for service in self._GetEnabledServices():
         if service in tostart:
            cmd = [service, "start"]
            if service in self._servicestates["upgradevibs"]:
               cmd.append("upgrade")
            else:
               cmd.append("install")
            try:
               RunCmdWithMsg(cmd)
            except Errors.InstallationError, e:
               failedservices.append(service)
               log.warn("Error in starting service '%s': %s" % (service, e))
      return failedservices

   def _StartJumpStartPlugins(self, plugins):
      cmd = ["/sbin/jumpstart", "--method=start",
             "--plugin=%s" % ",".join(plugins)]
      rc, out = runcommand.runcommand(cmd)
      if rc:
         log.warn("Error loading one or more JumpStart plugins: %s" % out)

   def _UpdateTriggers(self, adds, removes):
      '''Activate trigger for VIB installation or removal
      '''
      for vib in adds:
         for trigger in self._triggers:
            trigger.Match(vib)
      for vib in removes:
         for trigger in self._triggers:
            trigger.Match(vib)


class LiveImageInstaller(Installer):
   '''LiveImageInstaller is the Installer class to live install/remove VIBs for
      live system.

      Attributes:
         * database - A Database.Database instance of the live system
   '''
   installertype = "live"
   priority = 5
   SUPPORTED_VIBS = set([Vib.BaseVib.TYPE_BOOTBANK,])
   SUPPORTED_PAYLOADS = set([Vib.Payload.TYPE_TGZ, Vib.Payload.TYPE_VGZ])

   def __init__(self, root = '/'):
      self.liveimage = LiveImage(root)
      self.problems = list()

   @property
   def database(self):
      return self.liveimage.database

   @property
   def stagedatabase(self):
      return self.liveimage.stagedatabase

   def StartTransaction(self, imgprofile, imgstate = None, preparedest = True,
                        forcebootbank = False, **kwargs):
      """Initiates a new installation transaction. Calculate what actions
         need to be taken.

         This method only works on staging directory

         Parameters:
            * imgprofile  - The ImageProfile instance representing the
                            target set of VIBs for the new image
            * imgstate    - The state of current HostImage, one of IMGSTATE_*
            * preparedest - Boolean, if True, then prepare the destination.
                            Set to false for a "dry run", to avoid changing
                            the destination.
            * forcebootbank - Boolean, if True, skip install of live image
                              even if its eligible for live install
         Returns:
            A tuple (installs, removes, staged), installs and removes are list
            of VIB IDs for HostImage.Stage() to install to the destination and
            to remove from the destination, in order to make it compliant
            with imgprofile. If LiveImage has already staged the imgprofile,
            staged is True.
            If there is nothing to do, (None, None, False) is returned.
         Exceptions:
            InstallationError
      """
      # Skip if reboot required VIBs have been installed
      from vmware.esximage.HostImage import HostImage
      if forcebootbank:
         msg = 'Nothing to do for live install - live installation has been ' \
               'disabled.'
         self.problems.append(msg)
         log.debug(msg)
         return (None, None, False)

      if imgstate not in (HostImage.IMGSTATE_FRESH_BOOT,
            HostImage.IMGSTATE_LIVE_UPDATED):
         msg = ('Only reboot-required installations are possible right now as '
                'a reboot-required installation has been done.')
         self.problems.append(msg)
         log.debug(msg)
         return (None, None, False)

      if self.database.profile is None:
         msg = 'No ImageProfile found for live image'
         raise Errors.InstallationError(msg)

      imgprofile = imgprofile.Copy()
      unsupported = imgprofile.vibIDs - self.GetSupportedVibs(imgprofile.vibs)
      for vibid in unsupported:
         imgprofile.RemoveVib(vibid)

      staged = False
      if self.stagedatabase is not None:
         if self.stagedatabase.profile and \
               self.stagedatabase.profile.vibIDs == imgprofile.vibIDs:
            staged = True
         else:
            self.liveimage.Cleanup()

      adds, removes = imgprofile.Diff(self.database.profile)

      if staged:
         return (adds, removes, staged)

      problems = self._CheckTransaction(imgprofile, adds, removes)
      if problems:
         log.debug('The transaction is not supported for live install:\n%s' % (problems))
         self.problems = problems
         return (None, None, False)

      if preparedest and (removes or adds):
         self.liveimage.StartTransaction(imgprofile)

      return (adds, removes, staged)

   def OpenPayloadFile(self, vibid, payload, read = True, write = False):
      """Creates and returns a File-like object for either reading from
         or writing to a given payload.  One of read or write must be True, but
         ready and write cannot both be true.

         Parameters:
            * vibid   - The Vib id containing the payload
            * payload - The Vib.Payload instance to read or write
            * read    - Set to True to get a File object for reading
                        from the payload.
            * write   - Set to True to get a File object for writing
                        to the payload.
         Returns:
            A File-like object, must support read (for read), write (for
            write), close methods.
            None if the desired read/write is not supported.
         Exceptions:
            AssertionError    - neither read nor write is True, or both are true
            InstallationError - Cannot open file to write or read
      """
      Installer.OpenPayloadFile(self, vibid, payload, read, write)


      if read == True:
         if vibid not in self.database.profile.vibstates:
            msg = 'Could not locate VIB %s in LiveImageInstaller' % (vibid)
            raise Errors.InstallationError(None, msg)

         vibstate = self.database.profile.vibstates[vibid]
         if payload.name not in vibstate.payloads:
            msg = "Payload name '%s' of VIB %s not in LiveImage DB" % (
                  payload.name, vibid)
            raise Errors.InstallationError(None, msg)

         if (payload.payloadtype in self.SUPPORTED_PAYLOADS or \
             payload.payloadtype == Vib.Payload.TYPE_BOOT):

            filepath = os.path.join(self.liveimage.tardisksdir,
                  vibstate.payloads[payload.name])
            if os.path.isfile(filepath):
               #NOTE: No name and time info in the gzip stream
               return EsxGzip.GunzipFile(filepath, 'rb')
            else:
               msg = "Payload '%s' of VIB %s at '%s' is missing" % (
                     payload.name, vibid, filepath)
               raise Errors.InstallationError(None, msg)
         else:
            return None
      else:
         if payload.payloadtype not in self.SUPPORTED_PAYLOADS:
            log.debug("Payload %s of type '%s' in VIB '%s' is not supported by "
                      "LiveImageInstaller." % (payload.name, payload.payloadtype,
                         vibid))
            return None

         filepath = os.path.join(self.liveimage.stagedatadir, payload.localname)
         try:
            fp = EsxGzip.GunzipFile(filepath, 'wb')
         except EnvironmentError, e:
            msg = 'Can not open %s to write payload %s: %s' % (filepath,
                  payload.name, e)
            raise Errors.InstallationError([], msg)
         return fp

      return None

   def Cleanup(self):
      '''Cleans up the live image staging area.
      '''
      try:
         self.liveimage.Cleanup()
      except:
         pass

   def CompleteStage(self):
      """Complete the staging of live image by writing out the database and
         staged indicator to staging directory.

         Exceptions:
            InstallationError
      """
      self.liveimage.CompleteStage()

   def Remediate(self, checkmaintmode=True):
      """Live remove and install VIB payloads

         For each VIB to remove, shutdown scripts will be executed and payloads
         will be unmounted. For each VIB to install, payloads will be mounted
         and init scripts will be executed.

         Returns:
            A Boolean, always False, as a reboot is not needed.
         Exceptions:
            InstallationError -
            HostNotChanged    - If there is no staged ImageProfile
      """
      if self.liveimage.isstaged:
         self.liveimage.Remediate(checkmaintmode)
      else:
         msg = 'LiveImage is not yet staged, nothing to remediate.'
         raise Errors.HostNotChanged(msg)

      return False

   def _CheckTransaction(self, imageprofile, adds, removes):
      '''Check the transaction to see if there are any logical reasons that
         prevent live installation of the transaction.
            * For VIBs to be removed: require liveremoveok
            * For VIBs to be installed: require liveinstallok
            * No VIBs to be installed are overlaid by existing VIB
            * No VIBs to be removed are overlaid by existing VIB
      '''
      problems = []

      # Build a map from file path to the file states. File state is a bit map
      # to indicate which group the file belongs to.
      filestates = {}
      keepreg, keepoverlay, removereg, removeoverlay, addreg, addoverlay = \
         1, 2, 4, 8, 16, 32

      keeps = self.database.profile.vibIDs - set(removes)
      groups = ((keeps, self.database.vibs, (keepreg, keepoverlay)),
                (removes, self.database.vibs, (removereg, removeoverlay)),
                (adds, imageprofile.vibs, (addreg, addoverlay)))
      for vibids, vibs, flags in groups:
         for vibid in vibids:
            vib = vibs[vibid]
            for filepath in vib.filelist:
               if filepath == '' or filepath.endswith('/'):
                     continue

               filepath = PathUtils.CustomNormPath('/' + filepath)
               ind = vib.overlay and 1 or 0
               if filepath in filestates:
                  filestates[filepath] |= flags[ind]
               else:
                  filestates[filepath] = flags[ind]

      unsupported = {
         keepoverlay | removereg  : \
            "File to be removed is overlaid by existing VIB",
         keepoverlay | removereg | addreg : \
            "File to be removed/installed is overlaid by existing VIB",
         keepoverlay | addreg : \
            "File to be installed is overlaid by existing VIB",
         #XXX: visorFS to support resurrect?
         keepreg | removeoverlay : \
            "File to be removed overlays existing VIB",
         keepreg | removeoverlay | addoverlay : \
            "File to be removed/installed overlays existing VIB",
            }

      for filepath in filestates:
         if filestates[filepath] in unsupported:
            problem = "%s : %s" % (unsupported[filestates[filepath]], filepath)
            problems.append(problem)

      for vibid in adds:
         # liveinstallok must be True for new VIB
         if not imageprofile.vibs[vibid].liveinstallok:
            problem = 'VIB %s cannot be live installed.' % (vibid)
            problems.append(problem)

      for vibid in removes:
         # liveremoveok must be Ture for VIB to be removed
         if not self.database.vibs[vibid].liveremoveok:
            problem = 'VIB %s cannot be removed live.' % (vibid)
            problems.append(problem)

      return problems

   def GetSupportedVibTypes(self):
      # Support locker VIB for PXE host so live install/remove
      # will keep tools VIB untouched in the image profile.
      if HostInfo.IsPxeBooting():
         return self.SUPPORTED_VIBS | set([Vib.BaseVib.TYPE_LOCKER,])
      else:
         return self.SUPPORTED_VIBS

class PostInstTrigger(object):
   '''Base class for system wide post installation trigger action. It does not
      implement any action.

      Attributes:
         * NAME - The name of the trigger
         * matchingvibs - List of VIB objects which trigger the class action
   '''
   NAME = ''

   def __init__(self):
      self.matchingvibs = []

   def Match(self, vib):
      '''Check the properties of a VIB instance. Add the VIB instance to
         matchingvibs list if install/remove this vib will need to run trigger
         action.

         Parameters:
            * vib - A VIB instanace, which will be installed or removed.
      '''
      raise NotImplementedError('Must instantiate a subclass of PostInstTrigger')

   def Run(self):
      '''Fire trigger action. This should be invoked after VIBs have been live
         installed/removed.

         NOTE: sub-class needs to implement _run() mehtod or override Run
               method.
      '''
      if self.matchingvibs:
         log.info("Executing post inst trigger : '%s'" % (self.NAME))
         self._run()

class CIMOMRestartTrigger(PostInstTrigger):
   '''Trigger to request cimom restart.
   '''
   NAME = 'CIMOM Restart Trigger'

   def Match(self, vib):
      if getattr(vib, 'cimomrestart', False):
         self.matchingvibs.append(vib)
         log.info('%s is required to install/remove %s' % (self.NAME, vib.id))

   def _run(self):
      cimomflag = os.path.join('/', SFCBD_RESTART_MARKER)
      log.debug("Processing '%s' to create file '%s' to request sfcb "
            "restart." % (self.NAME, cimomflag))
      try:
         open(cimomflag, 'w').close()
      except IOError, e:
         log.warning("%s : Creating file failed: %s" % (
            self.TRIGGER_NAME, e))

class WSMANRestartTrigger(PostInstTrigger):
   """Restart wsman when configuration is updated."""
   NAME = "WSMAN Restart Trigger"

   def Match(self, vib):
      matchfn = lambda x: fnmatch.fnmatch(x, "etc/cim/openwsman/*.conf")
      if LiveImage._GetFilesFromVib(vib, matchfn):
         self.matchingvibs.append(vib)
         log.info("%s matched VIB %s" % (self.NAME, vib.id))

   def _run(self):
      cmd = "/etc/init.d/wsman restart"
      try:
         RunCmdWithMsg(cmd, raiseexception=False)
      except runcommand.RunCommandError, e:
         log.warning("Error running [%s]: %s" % (cmd, e))

class FirewallRefreshTrigger(PostInstTrigger):
   '''FirewallRefreshTrigger is the trigger to refresh firewall settings
   '''
   NAME = 'Firewall Refresh Trigger'

   def Match(self, vib):
      if LiveImage._GetFilesFromVib(vib,
            lambda x: fnmatch.fnmatch(x, "etc/vmware/firewall/*")):
         self.matchingvibs.append(vib)
         log.info('%s is required to install/remove %s' % (self.NAME, vib.id))

   def _run(self):
      cmds = ['export VI_USERNAME=vpxuser; /sbin/esxcli network firewall refresh',
              '/bin/vim-cmd -U vpxuser hostsvc/refresh_firewall']
      for cmd in cmds:
         try:
            RunCmdWithMsg(cmd, raiseexception=False)
         except runcommand.RunCommandError, e:
            log.warning('There was an error to run [%s]: %s' % (cmd, e))

class HostServiceRefreshTrigger(PostInstTrigger):
   '''HostServiceRefreshTrigger is the trigger to refresh host service system
   '''
   NAME = 'Service System Refresh Trigger'

   def Match(self, vib):
      if LiveImage._GetFilesFromVib(vib,
            lambda x: fnmatch.fnmatch(x, "etc/vmware/service/*")):
         self.matchingvibs.append(vib)
         log.info('%s is required to install/remove %s' % (self.NAME, vib.id))

   def _run(self):
      cmd = '/bin/vim-cmd -U vpxuser hostsvc/refresh_service'
      try:
         RunCmdWithMsg(cmd, raiseexception=False)
      except runcommand.RunCommandError, e:
         log.warning('There was an error to run [%s]: %s' % (cmd, e))

def RunCmdWithMsg(cmd, title='', raiseexception=True):
   if title:
      log.debug(title)
   else:
      log.debug('Running [%s]...' % (cmd))
   rc, out = runcommand.runcommand(cmd)
   if out:
      log.debug("output: %s" % (out))

   if rc != 0:
      msg = ('Error in running %s:\nReturn code: %s'
             '\nOutput: %s\n%s' % (cmd, rc, out, FAILURE_WARNING))
      if raiseexception:
         raise Errors.InstallationError([], msg)
      else:
         log.warning(msg)

