#!/usr/bin/python
########################################################################
# Copyright (C) 2010 VMWare, Inc.                                      #
# All Rights Reserved                                                  #
########################################################################

"Provides a class for parsing and writing boot.cfg"

PRINTABLE = (""" !"#$%&'()*+,-./0123456789:;<=>?@"""
             """ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`"""
             """abcdefghijklmnopqrstuvwxyz{|}~""")
SPACE = "\t\n\v\f\r "

class BootCfgError(Exception):
   pass

class BootCfg(object):
   """
      Attributes:
         * bootstate - An integer, one of 0, 1, 2, 3, or 4.
         * title     - A string specifying a title to show during boot.
         * kernel    - A string specifying the name of the kernel image. Note
                       that this is not always vmkernel. It could be a boot
                       module that gets loaded before vmkernel, such as tboot.
         * kernelopt - A dictionary of options to be passed to the kernel
                       command line. Keys must be strings. Values may be either
                       strings or None (to indicate a keyword-only argument).
                       Spaces are not valid.
         * modules   - A list of module names. Each item must be a string.
         * build     - A string.
         * updated   - A non-negative integer.
   """
   def __init__(self, f=None):
      """Class constructor.
            Parameters:
               * f - If specified, either a file name or a file-like object
                     from which to parse configuration.
            Raises:
               * IOError      - If file name or file object is not valid.
               * BootCfgError - If file format is not correct.
      """
      self.bootstate = 0
      self.title = ""
      self.kernel = ""
      self.kernelopt = dict()
      self.modules = list()
      self.build = ""
      self.updated = 0

      if f is not None:
         self.Parse(f)

   def Parse(self, f):
      """Read configuration from file.
            Parameters:
               * f - Either a file name or a file-like object from which to
                     parse configuration.
            Raises:
               * IOError      - If file name or file object is not valid.
               * BootCfgError - If file format is not correct.
      """
      if isinstance(f, basestring):
         fobj = open(f, "rb")
         fn = f
      else:
         fobj = f
         fn = hasattr(fobj, "name") and fobj.name or "<file>"

      lineno = 0
      for line in fobj:
         lineno += 1
         # Comments are not preserved.
         line = line.split("#")[0].strip()
         if not line:
            continue
         try:
            name, value = (word.strip() for word in line.split("=", 1))
         except ValueError:
            msg = "Invalid format at line %d of %s" % (lineno, fn)
            raise BootCfgError(msg)
         if name == "bootstate":
            try:
               self.bootstate = int(value)
            except ValueError:
               msg = "Invalid 'bootstate' value at line %d of %s" % (lineno, fn)
               raise BootCfgError(msg)
         elif name == "title":
            self.title = value
         elif name == "kernel":
            self.kernel = value
         elif name == "kernelopt":
            try:
               self._ParseKernelOpt(value)
            except IndexError:
               msg = "Invalid 'kernelopt' value at line %s of %s" % (lineno, fn)
               raise BootCfgError(msg)
         elif name == "modules":
            self.modules = [word.strip() for word in value.split(" --- ")]
         elif name == "build":
            self.build = value
         elif name == "updated":
            try:
               self.updated = int(value)
            except ValueError:
               msg = "Invalid 'updated' value at line %d of %s" % (lineno, fn)
               raise BootCfgError(msg)
         else:
            msg = "Unknown keyword '%s' at line %s of %s" % (name, lineno, fn)
            raise BootCfgError(msg)

   def _ParseKernelOpt(self, line):
      # A Pythonification of what's in bora/vmkernel/lib/bootConfigParse.c.
      i = 0
      linelen = len(line)
      while i < linelen:
         while line[i] in SPACE:
            i += 1
            if i == linelen:
               return

         if line[i] in "#\0\n":
            break

         keystart = i
         while i < linelen and line[i] not in " =:" and line[i] in PRINTABLE:
            i += 1
         key = line[keystart:i]
         self.kernelopt[key] = None

         if i == linelen:
            return

         while line[i] in SPACE:
            i += 1
            if i == linelen:
               return

         if line[i] not in "=:":
            continue

         i += 1
         if i == linelen:
            break

         while line[i] in SPACE:
            i += 1
            if i == linelen:
               return

         valstart = i
         while line[i] not in " =," and line[i] in PRINTABLE:
            i += 1
            if i == linelen:
               break
         self.kernelopt[key] = line[valstart:i].strip('"')

         if i == linelen:
            return

         while line[i] in SPACE:
            i += 1
            if i == linelen:
               return

         if line[i] == ",":
            i += 1
            if i == linelen:
               return

   def Write(self, f):
      """Write configuration to a file.
            Parameters:
               * f - Either a file name or a file-like object to which to
                     write configuration.
            Raises:
               * IOError      - If file name or file object is not valid.
      """
      if isinstance(f, basestring):
         fobj = open(f, "wb")
      else:
         fobj = f
      fobj.write("bootstate=")
      fobj.write(str(self.bootstate))
      fobj.write("\ntitle=")
      fobj.write(self.title)
      fobj.write("\nkernel=")
      fobj.write(self.kernel)
      fobj.write("\nkernelopt=")
      makeopt = lambda k, v: v is None and k or "=".join((k, v))
      fobj.write(" ".join(makeopt(k, v) for k, v in self.kernelopt.iteritems()))
      fobj.write("\nmodules=")
      fobj.write(" --- ".join(self.modules))
      fobj.write("\nbuild=")
      fobj.write(self.build)
      fobj.write("\nupdated=")
      fobj.write(str(self.updated))
      fobj.write("\n")
